
/***************************************************************************
NWNX Time - HashSetPlugin functions plugin
Copyright (C) 2006 Ingmar Stieger (Papillon, papillon@nwnx.org)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
***************************************************************************/

#include <wx/textfile.h>
#include "HashSetPlugin.h"
#include "../../misc/log.h"

/***************************************************************************
NWNX and DLL specific functions
***************************************************************************/

HashSetPlugin* plugin;

DLLEXPORT Plugin* GetPluginPointer()
{
	return plugin;
}

BOOL APIENTRY DllMain(HMODULE hModule, DWORD ul_reason_for_call, LPVOID lpReserved)
{
	if (ul_reason_for_call == DLL_PROCESS_ATTACH)
	{
		plugin = new HashSetPlugin();

		TCHAR szPath[MAX_PATH];
		GetModuleFileName(hModule, szPath, MAX_PATH);
		plugin->SetPluginFullPath(szPath);
	}
	else if (ul_reason_for_call == DLL_PROCESS_DETACH)
	{
		delete plugin;
	}
	return TRUE;
}


/***************************************************************************
Implementation of HashSetPlugin Plugin
***************************************************************************/

// TODO: If possible, this should be a configurable initial map size? 
//       m_maps will probably need to be a pointer to make that doable.
HashSetPlugin::HashSetPlugin() : 
	m_maps(250), 
	m_knownCommands(21) // 16 commands + 30%
{
	header = _T(
		"NWNX HashSet Plugin V.1.0.0\n" \
		"(c) 2007 by Marc Paradise [Grinning Fool]\n" \
		"visit us at http://www.khalidine.com\n");

	description = _T(
		"This plugin is a port of the original NWNX2 hashet plugin, along with some extra goodies. ");
	
	subClass = _T("HASHSET");
	version = _T("1.0.0");
	m_pData = NULL;
	m_pMapTokenizer = NULL;
	// I may be getting a bit obsessive with the maps here... 
	m_knownCommands["CREATE"] =	create;
	m_knownCommands["DESTROY"] = destroy;
	m_knownCommands["VALID"] = isValid;
	m_knownCommands["EXISTS"] = isExisting;
	m_knownCommands["INSERT"] = insert;
	m_knownCommands["LOOKUP"] = lookup;
	m_knownCommands["DELETE"] = del;
	m_knownCommands["GETFIRSTKEY"] = getFirstKey;
	m_knownCommands["GETNEXTKEY"] = getNextKey;
	m_knownCommands["HASNEXTKEY"] = hasNextKey;
	m_knownCommands["GETFIRSTVAL"] = getFirstValue;
	m_knownCommands["GETNEXTVAL"] = getNextValue;
	m_knownCommands["HASNEXTVAL"] = hasNextValue;
	m_knownCommands["GETSIZE"] = getSize;
	m_knownCommands["LOADFILE"] = loadFromFile;
	m_knownCommands["TOKMAPNEW"] = startMappedToken;
	m_knownCommands["TOKMAPNEXT"] = nextMappedToken;

}

HashSetPlugin::~HashSetPlugin()
{
	wxLogMessage(wxT("* Plugin unloaded."));
	if (m_pData) 
		delete [] m_pData;
	if (m_pMapTokenizer) 
		delete m_pMapTokenizer;
	HashMaps::iterator end = m_maps.end();
	HashMaps::iterator it = m_maps.begin();
	while (it++ != end) { 
		delete it->second;
	}
	m_maps.clear();
	
	
}

bool HashSetPlugin::Init(TCHAR* nwnxhome)
{
	assert(GetPluginFileName());

	/* Log file */
	wxString logfile(nwnxhome); 
	logfile.append(wxT("\\"));
	logfile.append(GetPluginFileName());
	logfile.append(wxT(".txt"));
	logger = new wxLogNWNX(logfile, wxString(header.c_str()));

	wxLogMessage(wxT("* Plugin initialized."));
	return true;
}

void HashSetPlugin::GetFunctionClass(TCHAR* fClass)
{
	_tcsncpy_s(fClass, 128, wxT("HASHSET"), 9); 
}

const char* HashSetPlugin::DoRequest(char *gameObject, char* request, char* parameters)
{
	//wxLogDebug(wxT("* Plugin DoRequest(0x%x, %s, %s)"), gameObject, request, parameters);
	wxLogMessage(wxT("* Plugin DoRequest(0x%x, %s, %s)"), gameObject, request, parameters);


#ifdef UNICODE
	wxString wxRequest(request, wxConvUTF8);
#else
	wxString wxRequest(request);
#endif
	wxString function;
	wxString mapName;
	wxString keyName;
	wxString data;
	HashData * map = NULL;
	wxStringTokenizer tkz(wxRequest, wxT("!"));
	m_returnData.Empty();
	
	// First param must always be function/action
	if (tkz.HasMoreTokens()) {
		function = tkz.GetNextToken();
	//	wxLogDebug(wxT("o Invoked Function: %s"), function);
		wxLogMessage(wxT("o Invoked Function: %s"), function);
	} else {
		wxLogError(wxT("o Function not specified."));
		return NULL;
	}
	CommandMap::iterator it = m_knownCommands.find(function);
	if (it == m_knownCommands.end()) { 
		wxLogError(wxT(" - Unknown function"));
		ProcessQueryFunction(function.c_str(), parameters);
		return NULL;	
	}
	// Second param is always map name.  We need make it specific to this object
	if (tkz.HasMoreTokens()) { 
		int objId = *(int*)(gameObject+4);
		if (objId == 0 || objId == 0xFF || objId == 0xFFFF || objId == 0xFFFFFF || objId == 0xFFFFFFFF ) { 
			wxLogError(wxT(" - Invalid OBJECTID"));
			return NULL;
		}
		mapName.Printf("%s-%d", tkz.GetNextToken(), objId);
		map = findMap(mapName);

		// Third param is usually the key or blank, save one condition	
		if (it->second != startMappedToken && tkz.HasMoreTokens()) { 
			keyName = tkz.GetNextToken();
		}
	}

	// These don't need a valid map -- checking them here first so I can avoid 
	// having to put a 'if (map != null)' and appropriate logging on all remaining 
	// options.
	switch (it->second) { 
		case create:  { // creating a new hashmap
			int size = atoi(keyName); 
			if (size == 0) { 
				wxLogError(wxT(" - no size specified in CREATE."));
				break;
			}
			wxLogInfo(wxT(" - Creating new hashmap under key '%s'"), mapName);
			deleteMap(mapName);
			m_maps[mapName] = new HashData(size);
			return NULL;
		}
		case isValid: // is the specified map name existing?
			m_returnData = map ? "1" : "0";
			return m_returnData.c_str();

		case loadFromFile:
			deleteMap(mapName);
			if (tkz.HasMoreTokens()) { 
				map = getMapFromFile(keyName, tkz.GetNextToken());
			} else { 
				wxLogError(wxT("- error, no separator character supplied!"));
				return NULL;
			}
			m_maps[mapName] = map;
			return NULL;
		default:
			if (map == NULL) { 
				wxLogError(" - map " + mapName + " not found for function " + function);
				return NULL;
			}
	}

	// TODO: Creating indexed array: can do this very simply as: 
	// receive new item, put instance of item in BOTH the hashmap and an array. 
	//  (resize array as needed) 
	// when performing 'get first/getnext' operations, or indexed operations, use the array
	// when performing key value lookups, use the hash. 
	// Allow option at creation to disable support fort he index/array option, if performance is a concenr. 
	switch (it->second) {

		case destroy: // destroy a hashmap
			deleteMap(mapName);
			break;


		case isExisting: // is the key within the map valid? 
			m_returnData = (keyName.IsEmpty() || !map || !map->getValue(keyName, data)) ? "0" : "1";
			break;

		case insert: // add a key-value to a map
			if (tkz.HasMoreTokens()) { 
				wxString x = tkz.GetNextToken();
				wxLogDebug(wxT(" - Adding map '%s', key '%s', value '%s'"), mapName, keyName, x);
				map->add(keyName, x);
			}
			break;

		case lookup:
			if (!map->getValue(keyName, m_returnData)) {
				wxLogDebug(wxT(" - key name not found: ") + keyName);
				m_returnData.Empty();
			}
			break;
			
		case del: 
			map->remove(keyName);
			break;

		case getSize:
			if (map) 
				m_returnData.Printf("%d", map->getCount());
			break;

		case startMappedToken:
			newTokenSearch(tkz.GetString());
			break;

		case nextMappedToken:
			nextTokenizerMatch(map, parameters);	
			wxLogDebug(wxT(" nextMappedToken - returning: '%s'") , parameters);
			return parameters;

		case getFirstKey: 
			map->getFirstKey(m_returnData);
			break;
		
		case hasNextKey:
			m_returnData = map->hasNextKey() ? "1" : "0";
			break;

		case getNextKey:
			map->getNextKey(m_returnData);
			break;
		case getFirstValue: 
			map->getFirstValue(m_returnData);
			break;
		
		case hasNextValue:
			m_returnData = map->hasNextValue() ? "1" : "0";
			break;

		case getNextValue:
			map->getNextValue(m_returnData);
			break;
		default:
			wxLogError(" - Error: Unknown command: %s", function);
			break;

	}
	return m_returnData.c_str();
	
}
// Safely delete a map entry, and remove any reference from the main hashmap
void HashSetPlugin::deleteMap(const wxString& name)
{
	HashData * m = findMap(name);
	if (m) { // If something already exists here, remove
		// it seems that sometimes the old pointer may be getting kept
		// so we're going to explicitly set it to NULL as well as invoking 'erase'
		m_maps[name] = NULL; 
		m_maps.erase(name);
		delete m;	
	}

}

HashData * HashSetPlugin::getMapFromFile(const wxString & filename, const wxString & sep)  
{
	wxLogMessage(wxT(" - Loading map data file %s"), filename);
#ifdef UNICODE
	wxTextFile file(filename, wxConvUTF8); 
#else
	wxTextFile file(filename);
#endif		

	if (!file.Exists()) { 
		wxLogError(wxT(" - file %s does not exist"), filename);
		return NULL;
	}
	if (!file.Open(filename)) { 
		wxLogError(wxT(" - file %s could not be opened."), filename);
		return NULL;
	}
	HashData * map = new HashData((int)file.GetLineCount() * 2);
	wxString str, value, key;
	for (str = file.GetFirstLine(); !file.Eof(); str = file.GetNextLine()) { 
		wxStringTokenizer tkz(str, wxT(sep));
		if (!tkz.HasMoreTokens()) { 
			continue;
		}
		key = tkz.GetNextToken();
		if (key.size() == 0) 
			continue;

		char c = key[0];
		if (c == ';')  
			continue;

		if (tkz.HasMoreTokens()) { 
			value = tkz.GetString();
		} else { 
			value = "";
		}
		
		wxLogDebug(wxT("Placing '%s' at '%s'"), value, key);
		map->add(key, value);
	}
	file.Close();
	return map;
}

HashData * HashSetPlugin::findMap(const wxString & name)
{
	if (name.IsEmpty())
		return NULL;
	HashMaps::iterator it = m_maps.find(name);
	if (it == m_maps.end()) {
		wxLogError(wxT("Could not find HashData instance: %s"), name);
		return NULL;
	}
	return it->second;
}

// Start a new search
void HashSetPlugin::newTokenSearch(const char * pszData) 
{
	if (m_pData) // reset our current token search walker.
		delete [] m_pData;
	if (m_pMapTokenizer) 
		delete m_pMapTokenizer;
	m_pData = new char[strlen(pszData) + 1];
	strcpy(m_pData, pszData);
	m_pMapTokenizer = new wxStringTokenizer(m_pData, ", ()[]?.!");
}
void HashSetPlugin::nextTokenizerMatch(HashData * map, char * output)
{
	if (!m_pMapTokenizer)  { 
		wxLogMessage(wxT(" -  no tokenizer exists!"));
		output[0] = 0;
		return;
	}
	wxString value;
	// Find the ntext token that occurs within the map
	while (m_pMapTokenizer->HasMoreTokens()) { 
		const char * p = m_pMapTokenizer->GetNextToken();
		wxLogDebug(wxT("Evaluating: %s"), p);
		if (map->getValue(p, value)) { 
			wxLogDebug(wxT("Match: %s => %s"), p, value);
			strcpy_s(output, strlen(output), value);
			return;
		}
		wxLogDebug(wxT("Looping for more tokens."));
		// Token not found. 
		continue;
	}
	// no more matches - done here.  Of course -- if the user has stored -1 
	// as a valid value, this is a problem; 
	wxLogDebug(wxT(" - out of matches, cleaning up tokenizer."));
	output[0] = 0;
	if (m_pData) { 
		delete []m_pData;
		m_pData = NULL;
	} 
	delete m_pMapTokenizer;
	m_pMapTokenizer = NULL;	
}



	


